@@ -1,23 +1,7 @@
 from __future__ import annotations
 
-import sys
-from typing import (
-    Any,
-    Dict,
-    Iterable,
-    Iterator,
-    List,
-    Mapping,
-    MutableMapping,
-    Tuple,
-    Union,
-)
-
-
-if sys.version_info[:2] >= (3, 8):
-    from typing import Protocol
-else:  # pragma: no cover
-    Protocol = object  # mypy will report errors on Python 3.7.
+from collections.abc import Iterable, Iterator, Mapping, MutableMapping
+from typing import Any, Protocol, Union
 
 
 __all__ = ["Headers", "HeadersLike", "MultipleValuesError"]
@@ -25,7 +9,7 @@ __all__ = ["Headers", "HeadersLike", "MultipleValuesError"]
 
 class MultipleValuesError(LookupError):
     """
-    Exception raised when :class:`Headers` has more than one value for a key.
+    Exception raised when :class:`Headers` has multiple values for a key.
 
     """
 
@@ -78,8 +62,8 @@ class Headers(MutableMapping[str, str]):
 
     # Like dict, Headers accepts an optional "mapping or iterable" argument.
     def __init__(self, *args: HeadersLike, **kwargs: str) -> None:
-        self._dict: Dict[str, List[str]] = {}
-        self._list: List[Tuple[str, str]] = []
+        self._dict: dict[str, list[str]] = {}
+        self._list: list[tuple[str, str]] = []
         self.update(*args, **kwargs)
 
     def __str__(self) -> str:
@@ -153,17 +137,17 @@ class Headers(MutableMapping[str, str]):
 
     # Methods for handling multiple values
 
-    def get_all(self, key: str) -> List[str]:
+    def get_all(self, key: str) -> list[str]:
         """
         Return the (possibly empty) list of all values for a header.
 
         Args:
-            key: header name.
+            key: Header name.
 
         """
         return self._dict.get(key.lower(), [])
 
-    def raw_items(self) -> Iterator[Tuple[str, str]]:
+    def raw_items(self) -> Iterator[tuple[str, str]]:
         """
         Return an iterator of all values as ``(name, value)`` pairs.
 
@@ -178,17 +162,16 @@ class SupportsKeysAndGetItem(Protocol):  # pragma: no cover
 
     """
 
-    def keys(self) -> Iterable[str]:
-        ...
+    def keys(self) -> Iterable[str]: ...
 
-    def __getitem__(self, key: str) -> str:
-        ...
+    def __getitem__(self, key: str) -> str: ...
 
 
+# Change to Headers | Mapping[str, str] | ... when dropping Python < 3.10.
 HeadersLike = Union[
     Headers,
     Mapping[str, str],
-    Iterable[Tuple[str, str]],
+    Iterable[tuple[str, str]],
     SupportsKeysAndGetItem,
 ]
 """
