@@ -1,38 +1,40 @@
-"""
-:mod:`websockets.legacy.client` defines the WebSocket client APIs.
-
-"""
+from __future__ import annotations
 
 import asyncio
-import collections.abc
 import functools
 import logging
+import os
+import random
+import traceback
+import urllib.parse
 import warnings
+from collections.abc import AsyncIterator, Generator, Sequence
 from types import TracebackType
-from typing import Any, Callable, Generator, List, Optional, Sequence, Tuple, Type, cast
+from typing import Any, Callable, cast
 
+from ..asyncio.compatibility import asyncio_timeout
 from ..datastructures import Headers, HeadersLike
 from ..exceptions import (
-    InvalidHandshake,
     InvalidHeader,
-    InvalidMessage,
-    InvalidStatusCode,
+    InvalidHeaderValue,
     NegotiationError,
-    RedirectHandshake,
     SecurityError,
 )
-from ..extensions.base import ClientExtensionFactory, Extension
+from ..extensions import ClientExtensionFactory, Extension
 from ..extensions.permessage_deflate import enable_client_permessage_deflate
 from ..headers import (
     build_authorization_basic,
     build_extension,
+    build_host,
     build_subprotocol,
     parse_extension,
     parse_subprotocol,
+    validate_subprotocols,
 )
-from ..http import USER_AGENT, build_host
-from ..typing import ExtensionHeader, Origin, Subprotocol
+from ..http11 import USER_AGENT
+from ..typing import ExtensionHeader, LoggerLike, Origin, Subprotocol
 from ..uri import WebSocketURI, parse_uri
+from .exceptions import InvalidMessage, InvalidStatusCode, RedirectHandshake
 from .handshake import build_request, check_response
 from .http import read_response
 from .protocol import WebSocketCommonProtocol
@@ -40,104 +42,31 @@ from .protocol import WebSocketCommonProtocol
 
 __all__ = ["connect", "unix_connect", "WebSocketClientProtocol"]
 
-logger = logging.getLogger("websockets.server")
-
 
 class WebSocketClientProtocol(WebSocketCommonProtocol):
     """
-    :class:`~asyncio.Protocol` subclass implementing a WebSocket client.
+    WebSocket client connection.
 
-    :class:`WebSocketClientProtocol`:
+    :class:`WebSocketClientProtocol` provides :meth:`recv` and :meth:`send`
+    coroutines for receiving and sending messages.
 
-    * performs the opening handshake to establish the connection;
-    * provides :meth:`recv` and :meth:`send` coroutines for receiving and
-      sending messages;
-    * deals with control frames automatically;
-    * performs the closing handshake to terminate the connection.
-
-    :class:`WebSocketClientProtocol` supports asynchronous iteration::
+    It supports asynchronous iteration to receive messages::
 
         async for message in websocket:
             await process(message)
 
-    The iterator yields incoming messages. It exits normally when the
-    connection is closed with the close code 1000 (OK) or 1001 (going away).
-    It raises a :exc:`~websockets.exceptions.ConnectionClosedError` exception
-    when the connection is closed with any other code.
+    The iterator exits normally when the connection is closed with close code
+    1000 (OK) or 1001 (going away) or without a close code. It raises
+    a :exc:`~websockets.exceptions.ConnectionClosedError` when the connection
+    is closed with any other code.
 
-    Once the connection is open, a `Ping frame`_ is sent every
-    ``ping_interval`` seconds. This serves as a keepalive. It helps keeping
-    the connection open, especially in the presence of proxies with short
-    timeouts on inactive connections. Set ``ping_interval`` to ``None`` to
-    disable this behavior.
+    See :func:`connect` for the documentation of ``logger``, ``origin``,
+    ``extensions``, ``subprotocols``, ``extra_headers``, and
+    ``user_agent_header``.
 
-    .. _Ping frame: https://tools.ietf.org/html/rfc6455#section-5.5.2
-
-    If the corresponding `Pong frame`_ isn't received within ``ping_timeout``
-    seconds, the connection is considered unusable and is closed with
-    code 1011. This ensures that the remote endpoint remains responsive. Set
-    ``ping_timeout`` to ``None`` to disable this behavior.
-
-    .. _Pong frame: https://tools.ietf.org/html/rfc6455#section-5.5.3
-
-    The ``close_timeout`` parameter defines a maximum wait time for completing
-    the closing handshake and terminating the TCP connection. For legacy
-    reasons, :meth:`close` completes in at most ``5 * close_timeout`` seconds.
-
-    ``close_timeout`` needs to be a parameter of the protocol because
-    websockets usually calls :meth:`close` implicitly upon exit when
-    :func:`connect` is used as a context manager.
-
-    To apply a timeout to any other API, wrap it in :func:`~asyncio.wait_for`.
-
-    The ``max_size`` parameter enforces the maximum size for incoming messages
-    in bytes. The default value is 1 MiB. ``None`` disables the limit. If a
-    message larger than the maximum size is received, :meth:`recv` will
-    raise :exc:`~websockets.exceptions.ConnectionClosedError` and the
-    connection will be closed with code 1009.
-
-    The ``max_queue`` parameter sets the maximum length of the queue that
-    holds incoming messages. The default value is ``32``. ``None`` disables
-    the limit. Messages are added to an in-memory queue when they're received;
-    then :meth:`recv` pops from that queue. In order to prevent excessive
-    memory consumption when messages are received faster than they can be
-    processed, the queue must be bounded. If the queue fills up, the protocol
-    stops processing incoming data until :meth:`recv` is called. In this
-    situation, various receive buffers (at least in :mod:`asyncio` and in the
-    OS) will fill up, then the TCP receive window will shrink, slowing down
-    transmission to avoid packet loss.
-
-    Since Python can use up to 4 bytes of memory to represent a single
-    character, each connection may use up to ``4 * max_size * max_queue``
-    bytes of memory to store incoming messages. By default, this is 128 MiB.
-    You may want to lower the limits, depending on your application's
-    requirements.
-
-    The ``read_limit`` argument sets the high-water limit of the buffer for
-    incoming bytes. The low-water limit is half the high-water limit. The
-    default value is 64 KiB, half of asyncio's default (based on the current
-    implementation of :class:`~asyncio.StreamReader`).
-
-    The ``write_limit`` argument sets the high-water limit of the buffer for
-    outgoing bytes. The low-water limit is a quarter of the high-water limit.
-    The default value is 64 KiB, equal to asyncio's default (based on the
-    current implementation of ``FlowControlMixin``).
-
-    As soon as the HTTP request and response in the opening handshake are
-    processed:
-
-    * the request path is available in the :attr:`path` attribute;
-    * the request and response HTTP headers are available in the
-      :attr:`request_headers` and :attr:`response_headers` attributes,
-      which are :class:`~websockets.http.Headers` instances.
-
-    If a subprotocol was negotiated, it's available in the :attr:`subprotocol`
-    attribute.
-
-    Once the connection is closed, the code is available in the
-    :attr:`close_code` attribute and the reason in :attr:`close_reason`.
-
-    All attributes must be treated as read-only.
+    See :class:`~websockets.legacy.protocol.WebSocketCommonProtocol` for the
+    documentation of ``ping_interval``, ``ping_timeout``, ``close_timeout``,
+    ``max_size``, ``max_queue``, ``read_limit``, and ``write_limit``.
 
     """
 
@@ -147,17 +76,22 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
     def __init__(
         self,
         *,
-        origin: Optional[Origin] = None,
-        extensions: Optional[Sequence[ClientExtensionFactory]] = None,
-        subprotocols: Optional[Sequence[Subprotocol]] = None,
-        extra_headers: Optional[HeadersLike] = None,
+        logger: LoggerLike | None = None,
+        origin: Origin | None = None,
+        extensions: Sequence[ClientExtensionFactory] | None = None,
+        subprotocols: Sequence[Subprotocol] | None = None,
+        extra_headers: HeadersLike | None = None,
+        user_agent_header: str | None = USER_AGENT,
         **kwargs: Any,
     ) -> None:
+        if logger is None:
+            logger = logging.getLogger("websockets.client")
+        super().__init__(logger=logger, **kwargs)
         self.origin = origin
         self.available_extensions = extensions
         self.available_subprotocols = subprotocols
         self.extra_headers = extra_headers
-        super().__init__(**kwargs)
+        self.user_agent_header = user_agent_header
 
     def write_http_request(self, path: str, headers: Headers) -> None:
         """
@@ -167,8 +101,10 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
         self.path = path
         self.request_headers = headers
 
-        logger.debug("%s > GET %s HTTP/1.1", self.side, path)
-        logger.debug("%s > %r", self.side, headers)
+        if self.debug:
+            self.logger.debug("> GET %s HTTP/1.1", path)
+            for key, value in headers.raw_items():
+                self.logger.debug("> %s: %s", key, value)
 
         # Since the path and headers only contain ASCII characters,
         # we can keep this simple.
@@ -177,28 +113,27 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
 
         self.transport.write(request.encode())
 
-    async def read_http_response(self) -> Tuple[int, Headers]:
+    async def read_http_response(self) -> tuple[int, Headers]:
         """
         Read status line and headers from the HTTP response.
 
         If the response contains a body, it may be read from ``self.reader``
         after this coroutine returns.
 
-        :raises ~websockets.exceptions.InvalidMessage: if the HTTP message is
-            malformed or isn't an HTTP/1.1 GET response
+        Raises:
+            InvalidMessage: If the HTTP message is malformed or isn't an
+                HTTP/1.1 GET response.
 
         """
         try:
             status_code, reason, headers = await read_response(self.reader)
-        # Remove this branch when dropping support for Python < 3.8
-        # because CancelledError no longer inherits Exception.
-        except asyncio.CancelledError:  # pragma: no cover
-            raise
         except Exception as exc:
             raise InvalidMessage("did not receive a valid HTTP response") from exc
 
-        logger.debug("%s < HTTP/1.1 %d %s", self.side, status_code, reason)
-        logger.debug("%s < %r", self.side, headers)
+        if self.debug:
+            self.logger.debug("< HTTP/1.1 %d %s", status_code, reason)
+            for key, value in headers.raw_items():
+                self.logger.debug("< %s: %s", key, value)
 
         self.response_headers = headers
 
@@ -207,8 +142,8 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
     @staticmethod
     def process_extensions(
         headers: Headers,
-        available_extensions: Optional[Sequence[ClientExtensionFactory]],
-    ) -> List[Extension]:
+        available_extensions: Sequence[ClientExtensionFactory] | None,
+    ) -> list[Extension]:
         """
         Handle the Sec-WebSocket-Extensions HTTP response header.
 
@@ -235,23 +170,20 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
         order of extensions, may be implemented by overriding this method.
 
         """
-        accepted_extensions: List[Extension] = []
+        accepted_extensions: list[Extension] = []
 
         header_values = headers.get_all("Sec-WebSocket-Extensions")
 
         if header_values:
-
             if available_extensions is None:
-                raise InvalidHandshake("no extensions supported")
+                raise NegotiationError("no extensions supported")
 
-            parsed_header_values: List[ExtensionHeader] = sum(
+            parsed_header_values: list[ExtensionHeader] = sum(
                 [parse_extension(header_value) for header_value in header_values], []
             )
 
             for name, response_params in parsed_header_values:
-
                 for extension_factory in available_extensions:
-
                     # Skip non-matching extensions based on their name.
                     if extension_factory.name != name:
                         continue
@@ -282,8 +214,8 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
 
     @staticmethod
     def process_subprotocol(
-        headers: Headers, available_subprotocols: Optional[Sequence[Subprotocol]]
-    ) -> Optional[Subprotocol]:
+        headers: Headers, available_subprotocols: Sequence[Subprotocol] | None
+    ) -> Subprotocol | None:
         """
         Handle the Sec-WebSocket-Protocol HTTP response header.
 
@@ -292,22 +224,23 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
         Return the selected subprotocol.
 
         """
-        subprotocol: Optional[Subprotocol] = None
+        subprotocol: Subprotocol | None = None
 
         header_values = headers.get_all("Sec-WebSocket-Protocol")
 
         if header_values:
-
             if available_subprotocols is None:
-                raise InvalidHandshake("no subprotocols supported")
+                raise NegotiationError("no subprotocols supported")
 
             parsed_header_values: Sequence[Subprotocol] = sum(
                 [parse_subprotocol(header_value) for header_value in header_values], []
             )
 
             if len(parsed_header_values) > 1:
-                subprotocols = ", ".join(parsed_header_values)
-                raise InvalidHandshake(f"multiple subprotocols: {subprotocols}")
+                raise InvalidHeaderValue(
+                    "Sec-WebSocket-Protocol",
+                    f"multiple values: {', '.join(parsed_header_values)}",
+                )
 
             subprotocol = parsed_header_values[0]
 
@@ -319,25 +252,25 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
     async def handshake(
         self,
         wsuri: WebSocketURI,
-        origin: Optional[Origin] = None,
-        available_extensions: Optional[Sequence[ClientExtensionFactory]] = None,
-        available_subprotocols: Optional[Sequence[Subprotocol]] = None,
-        extra_headers: Optional[HeadersLike] = None,
+        origin: Origin | None = None,
+        available_extensions: Sequence[ClientExtensionFactory] | None = None,
+        available_subprotocols: Sequence[Subprotocol] | None = None,
+        extra_headers: HeadersLike | None = None,
     ) -> None:
         """
         Perform the client side of the opening handshake.
 
-        :param origin: sets the Origin HTTP header
-        :param available_extensions: list of supported extensions in the order
-            in which they should be used
-        :param available_subprotocols: list of supported subprotocols in order
-            of decreasing preference
-        :param extra_headers: sets additional HTTP request headers; it must be
-            a :class:`~websockets.http.Headers` instance, a
-            :class:`~collections.abc.Mapping`, or an iterable of ``(name,
-            value)`` pairs
-        :raises ~websockets.exceptions.InvalidHandshake: if the handshake
-            fails
+        Args:
+            wsuri: URI of the WebSocket server.
+            origin: Value of the ``Origin`` header.
+            extensions: List of supported extensions, in order in which they
+                should be negotiated and run.
+            subprotocols: List of supported subprotocols, in order of decreasing
+                preference.
+            extra_headers: Arbitrary HTTP headers to add to the handshake request.
+
+        Raises:
+            InvalidHandshake: If the handshake fails.
 
         """
         request_headers = Headers()
@@ -367,15 +300,11 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
             protocol_header = build_subprotocol(available_subprotocols)
             request_headers["Sec-WebSocket-Protocol"] = protocol_header
 
-        if extra_headers is not None:
-            if isinstance(extra_headers, Headers):
-                extra_headers = extra_headers.raw_items()
-            elif isinstance(extra_headers, collections.abc.Mapping):
-                extra_headers = extra_headers.items()
-            for name, value in extra_headers:
-                request_headers[name] = value
+        if self.extra_headers is not None:
+            request_headers.update(self.extra_headers)
 
-        request_headers.setdefault("User-Agent", USER_AGENT)
+        if self.user_agent_header:
+            request_headers.setdefault("User-Agent", self.user_agent_header)
 
         self.write_http_request(wsuri.resource_name, request_headers)
 
@@ -385,7 +314,7 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
                 raise InvalidHeader("Location")
             raise RedirectHandshake(response_headers["Location"])
         elif status_code != 101:
-            raise InvalidStatusCode(status_code)
+            raise InvalidStatusCode(status_code, response_headers)
 
         check_response(response_headers, key)
 
@@ -402,86 +331,114 @@ class WebSocketClientProtocol(WebSocketCommonProtocol):
 
 class Connect:
     """
-    Connect to the WebSocket server at the given ``uri``.
+    Connect to the WebSocket server at ``uri``.
 
     Awaiting :func:`connect` yields a :class:`WebSocketClientProtocol` which
     can then be used to send and receive messages.
 
-    :func:`connect` can also be used as a asynchronous context manager::
+    :func:`connect` can be used as a asynchronous context manager::
 
         async with connect(...) as websocket:
             ...
 
-    In that case, the connection is closed when exiting the context.
+    The connection is closed automatically when exiting the context.
 
-    :func:`connect` is a wrapper around the event loop's
-    :meth:`~asyncio.loop.create_connection` method. Unknown keyword arguments
-    are passed to :meth:`~asyncio.loop.create_connection`.
+    :func:`connect` can be used as an infinite asynchronous iterator to
+    reconnect automatically on errors::
 
-    For example, you can set the ``ssl`` keyword argument to a
-    :class:`~ssl.SSLContext` to enforce some TLS settings. When connecting to
-    a ``wss://`` URI, if this argument isn't provided explicitly,
-    :func:`ssl.create_default_context` is called to create a context.
+        async for websocket in connect(...):
+            try:
+                ...
+            except websockets.exceptions.ConnectionClosed:
+                continue
 
-    You can connect to a different host and port from those found in ``uri``
-    by setting ``host`` and ``port`` keyword arguments. This only changes the
-    destination of the TCP connection. The host name from ``uri`` is still
-    used in the TLS handshake for secure connections and in the ``Host`` HTTP
-    header.
+    The connection is closed automatically after each iteration of the loop.
 
-    ``create_protocol`` defaults to :class:`WebSocketClientProtocol`. It may
-    be replaced by a wrapper or a subclass to customize the protocol that
-    manages the connection.
+    If an error occurs while establishing the connection, :func:`connect`
+    retries with exponential backoff. The backoff delay starts at three
+    seconds and increases up to one minute.
 
-    The behavior of ``ping_interval``, ``ping_timeout``, ``close_timeout``,
-    ``max_size``, ``max_queue``, ``read_limit``, and ``write_limit`` is
-    described in :class:`WebSocketClientProtocol`.
+    If an error occurs in the body of the loop, you can handle the exception
+    and :func:`connect` will reconnect with the next iteration; or you can
+    let the exception bubble up and break out of the loop. This lets you
+    decide which errors trigger a reconnection and which errors are fatal.
 
-    :func:`connect` also accepts the following optional arguments:
+    Args:
+        uri: URI of the WebSocket server.
+        create_protocol: Factory for the :class:`asyncio.Protocol` managing
+            the connection. It defaults to :class:`WebSocketClientProtocol`.
+            Set it to a wrapper or a subclass to customize connection handling.
+        logger: Logger for this client.
+            It defaults to ``logging.getLogger("websockets.client")``.
+            See the :doc:`logging guide <../../topics/logging>` for details.
+        compression: The "permessage-deflate" extension is enabled by default.
+            Set ``compression`` to :obj:`None` to disable it. See the
+            :doc:`compression guide <../../topics/compression>` for details.
+        origin: Value of the ``Origin`` header, for servers that require it.
+        extensions: List of supported extensions, in order in which they
+            should be negotiated and run.
+        subprotocols: List of supported subprotocols, in order of decreasing
+            preference.
+        extra_headers: Arbitrary HTTP headers to add to the handshake request.
+        user_agent_header: Value of  the ``User-Agent`` request header.
+            It defaults to ``"Python/x.y.z websockets/X.Y"``.
+            Setting it to :obj:`None` removes the header.
+        open_timeout: Timeout for opening the connection in seconds.
+            :obj:`None` disables the timeout.
 
-    * ``compression`` is a shortcut to configure compression extensions;
-      by default it enables the "permessage-deflate" extension; set it to
-      ``None`` to disable compression.
-    * ``origin`` sets the Origin HTTP header.
-    * ``extensions`` is a list of supported extensions in order of
-      decreasing preference.
-    * ``subprotocols`` is a list of supported subprotocols in order of
-      decreasing preference.
-    * ``extra_headers`` sets additional HTTP request headers; it can be a
-      :class:`~websockets.http.Headers` instance, a
-      :class:`~collections.abc.Mapping`, or an iterable of ``(name, value)``
-      pairs.
+    See :class:`~websockets.legacy.protocol.WebSocketCommonProtocol` for the
+    documentation of ``ping_interval``, ``ping_timeout``, ``close_timeout``,
+    ``max_size``, ``max_queue``, ``read_limit``, and ``write_limit``.
 
-    :raises ~websockets.uri.InvalidURI: if ``uri`` is invalid
-    :raises ~websockets.handshake.InvalidHandshake: if the opening handshake
-        fails
+    Any other keyword arguments are passed the event loop's
+    :meth:`~asyncio.loop.create_connection` method.
+
+    For example:
+
+    * You can set ``ssl`` to a :class:`~ssl.SSLContext` to enforce TLS
+      settings. When connecting to a ``wss://`` URI, if ``ssl`` isn't
+      provided, a TLS context is created
+      with :func:`~ssl.create_default_context`.
+
+    * You can set ``host`` and ``port`` to connect to a different host and
+      port from those found in ``uri``. This only changes the destination of
+      the TCP connection. The host name from ``uri`` is still used in the TLS
+      handshake for secure connections and in the ``Host`` header.
+
+    Raises:
+        InvalidURI: If ``uri`` isn't a valid WebSocket URI.
+        OSError: If the TCP connection fails.
+        InvalidHandshake: If the opening handshake fails.
+        ~asyncio.TimeoutError: If the opening handshake times out.
 
     """
 
-    MAX_REDIRECTS_ALLOWED = 10
+    MAX_REDIRECTS_ALLOWED = int(os.environ.get("WEBSOCKETS_MAX_REDIRECTS", "10"))
 
     def __init__(
         self,
         uri: str,
         *,
-        create_protocol: Optional[Callable[[Any], WebSocketClientProtocol]] = None,
-        ping_interval: Optional[float] = 20,
-        ping_timeout: Optional[float] = 20,
-        close_timeout: Optional[float] = None,
-        max_size: Optional[int] = 2 ** 20,
-        max_queue: Optional[int] = 2 ** 5,
-        read_limit: int = 2 ** 16,
-        write_limit: int = 2 ** 16,
-        loop: Optional[asyncio.AbstractEventLoop] = None,
-        compression: Optional[str] = "deflate",
-        origin: Optional[Origin] = None,
-        extensions: Optional[Sequence[ClientExtensionFactory]] = None,
-        subprotocols: Optional[Sequence[Subprotocol]] = None,
-        extra_headers: Optional[HeadersLike] = None,
+        create_protocol: Callable[..., WebSocketClientProtocol] | None = None,
+        logger: LoggerLike | None = None,
+        compression: str | None = "deflate",
+        origin: Origin | None = None,
+        extensions: Sequence[ClientExtensionFactory] | None = None,
+        subprotocols: Sequence[Subprotocol] | None = None,
+        extra_headers: HeadersLike | None = None,
+        user_agent_header: str | None = USER_AGENT,
+        open_timeout: float | None = 10,
+        ping_interval: float | None = 20,
+        ping_timeout: float | None = 20,
+        close_timeout: float | None = None,
+        max_size: int | None = 2**20,
+        max_queue: int | None = 2**5,
+        read_limit: int = 2**16,
+        write_limit: int = 2**16,
         **kwargs: Any,
     ) -> None:
         # Backwards compatibility: close_timeout used to be called timeout.
-        timeout: Optional[float] = kwargs.pop("timeout", None)
+        timeout: float | None = kwargs.pop("timeout", None)
         if timeout is None:
             timeout = 10
         else:
@@ -491,7 +448,7 @@ class Connect:
             close_timeout = timeout
 
         # Backwards compatibility: create_protocol used to be called klass.
-        klass: Optional[Type[WebSocketClientProtocol]] = kwargs.pop("klass", None)
+        klass: type[WebSocketClientProtocol] | None = kwargs.pop("klass", None)
         if klass is None:
             klass = WebSocketClientProtocol
         else:
@@ -503,8 +460,13 @@ class Connect:
         # Backwards compatibility: recv() used to return None on closed connections
         legacy_recv: bool = kwargs.pop("legacy_recv", False)
 
-        if loop is None:
+        # Backwards compatibility: the loop parameter used to be supported.
+        _loop: asyncio.AbstractEventLoop | None = kwargs.pop("loop", None)
+        if _loop is None:
             loop = asyncio.get_event_loop()
+        else:
+            loop = _loop
+            warnings.warn("remove loop argument", DeprecationWarning)
 
         wsuri = parse_uri(uri)
         if wsuri.secure:
@@ -520,8 +482,20 @@ class Connect:
         elif compression is not None:
             raise ValueError(f"unsupported compression: {compression}")
 
+        if subprotocols is not None:
+            validate_subprotocols(subprotocols)
+
+        # Help mypy and avoid this error: "type[WebSocketClientProtocol] |
+        # Callable[..., WebSocketClientProtocol]" not callable  [misc]
+        create_protocol = cast(Callable[..., WebSocketClientProtocol], create_protocol)
         factory = functools.partial(
             create_protocol,
+            logger=logger,
+            origin=origin,
+            extensions=extensions,
+            subprotocols=subprotocols,
+            extra_headers=extra_headers,
+            user_agent_header=user_agent_header,
             ping_interval=ping_interval,
             ping_timeout=ping_timeout,
             close_timeout=close_timeout,
@@ -529,30 +503,28 @@ class Connect:
             max_queue=max_queue,
             read_limit=read_limit,
             write_limit=write_limit,
-            loop=loop,
             host=wsuri.host,
             port=wsuri.port,
             secure=wsuri.secure,
             legacy_recv=legacy_recv,
-            origin=origin,
-            extensions=extensions,
-            subprotocols=subprotocols,
-            extra_headers=extra_headers,
+            loop=_loop,
         )
 
         if kwargs.pop("unix", False):
-            path: Optional[str] = kwargs.pop("path", None)
+            path: str | None = kwargs.pop("path", None)
             create_connection = functools.partial(
                 loop.create_unix_connection, factory, path, **kwargs
             )
         else:
-            host: Optional[str]
-            port: Optional[int]
+            host: str | None
+            port: int | None
             if kwargs.get("sock") is None:
                 host, port = wsuri.host, wsuri.port
             else:
                 # If sock is given, host and port shouldn't be specified.
                 host, port = None, None
+                if kwargs.get("ssl"):
+                    kwargs.setdefault("server_hostname", wsuri.host)
             # If host and port are given, override values from the URI.
             host = kwargs.pop("host", host)
             port = kwargs.pop("port", port)
@@ -560,35 +532,49 @@ class Connect:
                 loop.create_connection, factory, host, port, **kwargs
             )
 
+        self.open_timeout = open_timeout
+        if logger is None:
+            logger = logging.getLogger("websockets.client")
+        self.logger = logger
+
         # This is a coroutine function.
         self._create_connection = create_connection
+        self._uri = uri
         self._wsuri = wsuri
 
     def handle_redirect(self, uri: str) -> None:
         # Update the state of this instance to connect to a new URI.
+        old_uri = self._uri
         old_wsuri = self._wsuri
-        new_wsuri = parse_uri(uri)
+        new_uri = urllib.parse.urljoin(old_uri, uri)
+        new_wsuri = parse_uri(new_uri)
 
         # Forbid TLS downgrade.
         if old_wsuri.secure and not new_wsuri.secure:
             raise SecurityError("redirect from WSS to WS")
 
         same_origin = (
-            old_wsuri.host == new_wsuri.host and old_wsuri.port == new_wsuri.port
+            old_wsuri.secure == new_wsuri.secure
+            and old_wsuri.host == new_wsuri.host
+            and old_wsuri.port == new_wsuri.port
         )
 
-        # Rewrite the host and port arguments for cross-origin redirects.
+        # Rewrite secure, host, and port for cross-origin redirects.
         # This preserves connection overrides with the host and port
         # arguments if the redirect points to the same host and port.
         if not same_origin:
-            # Replace the host and port argument passed to the protocol factory.
             factory = self._create_connection.args[0]
+            # Support TLS upgrade.
+            if not old_wsuri.secure and new_wsuri.secure:
+                factory.keywords["secure"] = True
+                self._create_connection.keywords.setdefault("ssl", True)
+            # Replace secure, host, and port arguments of the protocol factory.
             factory = functools.partial(
                 factory.func,
                 *factory.args,
                 **dict(factory.keywords, host=new_wsuri.host, port=new_wsuri.port),
             )
-            # Replace the host and port argument passed to create_connection.
+            # Replace secure, host, and port arguments of create_connection.
             self._create_connection = functools.partial(
                 self._create_connection.func,
                 *(factory, new_wsuri.host, new_wsuri.port),
@@ -596,35 +582,73 @@ class Connect:
             )
 
         # Set the new WebSocket URI. This suffices for same-origin redirects.
+        self._uri = new_uri
         self._wsuri = new_wsuri
 
-    # async with connect(...)
+    # async for ... in connect(...):
+
+    BACKOFF_INITIAL = float(os.environ.get("WEBSOCKETS_BACKOFF_INITIAL_DELAY", "5"))
+    BACKOFF_MIN = float(os.environ.get("WEBSOCKETS_BACKOFF_MIN_DELAY", "3.1"))
+    BACKOFF_MAX = float(os.environ.get("WEBSOCKETS_BACKOFF_MAX_DELAY", "90.0"))
+    BACKOFF_FACTOR = float(os.environ.get("WEBSOCKETS_BACKOFF_FACTOR", "1.618"))
+
+    async def __aiter__(self) -> AsyncIterator[WebSocketClientProtocol]:
+        backoff_delay = self.BACKOFF_MIN / self.BACKOFF_FACTOR
+        while True:
+            try:
+                async with self as protocol:
+                    yield protocol
+            except Exception as exc:
+                # Add a random initial delay between 0 and 5 seconds.
+                # See 7.2.3. Recovering from Abnormal Closure in RFC 6455.
+                if backoff_delay == self.BACKOFF_MIN:
+                    initial_delay = random.random() * self.BACKOFF_INITIAL
+                    self.logger.info(
+                        "connect failed; reconnecting in %.1f seconds: %s",
+                        initial_delay,
+                        # Remove first argument when dropping Python 3.9.
+                        traceback.format_exception_only(type(exc), exc)[0].strip(),
+                    )
+                    await asyncio.sleep(initial_delay)
+                else:
+                    self.logger.info(
+                        "connect failed again; retrying in %d seconds: %s",
+                        int(backoff_delay),
+                        # Remove first argument when dropping Python 3.9.
+                        traceback.format_exception_only(type(exc), exc)[0].strip(),
+                    )
+                    await asyncio.sleep(int(backoff_delay))
+                # Increase delay with truncated exponential backoff.
+                backoff_delay = backoff_delay * self.BACKOFF_FACTOR
+                backoff_delay = min(backoff_delay, self.BACKOFF_MAX)
+                continue
+            else:
+                # Connection succeeded - reset backoff delay
+                backoff_delay = self.BACKOFF_MIN
+
+    # async with connect(...) as ...:
 
     async def __aenter__(self) -> WebSocketClientProtocol:
         return await self
 
     async def __aexit__(
         self,
-        exc_type: Optional[Type[BaseException]],
-        exc_value: Optional[BaseException],
-        traceback: Optional[TracebackType],
+        exc_type: type[BaseException] | None,
+        exc_value: BaseException | None,
+        traceback: TracebackType | None,
     ) -> None:
         await self.protocol.close()
 
-    # await connect(...)
+    # ... = await connect(...)
 
     def __await__(self) -> Generator[Any, None, WebSocketClientProtocol]:
         # Create a suitable iterator by calling __await__ on a coroutine.
         return self.__await_impl__().__await__()
 
     async def __await_impl__(self) -> WebSocketClientProtocol:
-        for redirects in range(self.MAX_REDIRECTS_ALLOWED):
-            transport, protocol = await self._create_connection()
-            # https://github.com/python/typeshed/pull/2756
-            transport = cast(asyncio.Transport, transport)
-            protocol = cast(WebSocketClientProtocol, protocol)
-
-            try:
+        async with asyncio_timeout(self.open_timeout):
+            for _redirects in range(self.MAX_REDIRECTS_ALLOWED):
+                _transport, protocol = await self._create_connection()
                 try:
                     await protocol.handshake(
                         self._wsuri,
@@ -633,19 +657,22 @@ class Connect:
                         available_subprotocols=protocol.available_subprotocols,
                         extra_headers=protocol.extra_headers,
                     )
-                except Exception:
+                except RedirectHandshake as exc:
+                    protocol.fail_connection()
+                    await protocol.wait_closed()
+                    self.handle_redirect(exc.uri)
+                # Avoid leaking a connected socket when the handshake fails.
+                except (Exception, asyncio.CancelledError):
                     protocol.fail_connection()
                     await protocol.wait_closed()
                     raise
                 else:
                     self.protocol = protocol
                     return protocol
-            except RedirectHandshake as exc:
-                self.handle_redirect(exc.uri)
-        else:
-            raise SecurityError("too many redirects")
+            else:
+                raise SecurityError("too many redirects")
 
-    # yield from connect(...)
+    # ... = yield from connect(...) - remove when dropping Python < 3.10
 
     __iter__ = __await__
 
@@ -654,20 +681,24 @@ connect = Connect
 
 
 def unix_connect(
-    path: Optional[str], uri: str = "ws://localhost/", **kwargs: Any
+    path: str | None = None,
+    uri: str = "ws://localhost/",
+    **kwargs: Any,
 ) -> Connect:
     """
     Similar to :func:`connect`, but for connecting to a Unix socket.
 
-    This function calls the event loop's
+    This function builds upon the event loop's
     :meth:`~asyncio.loop.create_unix_connection` method.
 
     It is only available on Unix.
 
     It's mainly useful for debugging servers listening on Unix sockets.
 
-    :param path: file system path to the Unix socket
-    :param uri: WebSocket URI
+    Args:
+        path: File system path to the Unix socket.
+        uri: URI of the WebSocket server; the host is used in the TLS
+            handshake for secure connections and in the ``Host`` header.
 
     """
     return connect(uri=uri, path=path, unix=True, **kwargs)
